import requests

from silence.settings import settings
from silence.logging.default_logger import logger
from silence import __version__

def check_for_new_version():
    if not settings.CHECK_FOR_UPDATES: return False

    logger.debug("Checking for new updates...")

    try:
        data = requests.get("https://pypi.org/pypi/Silence/json").json()
        latest_version = data["info"]["version"]
    except Exception as exc:
        logger.debug("Exception occurred when checking for updates (muted):")
        logger.debug(str(exc))
        return False

    current_version = __version__.split("-")[0]
    if current_version != latest_version:
        logger.debug("New version available")
        return latest_version
    else:
        logger.debug("Running latest version")
        return False



# https://stackoverflow.com/questions/12971631/sorting-list-by-an-attribute-that-can-be-none
from functools import total_ordering

# An object that compares as less than anything else regardless of their type
@total_ordering
class MinType(object):
    
    def __le__(self, other):
        return True

    def __eq__(self, other):
        return (self is other)

Min = MinType()



from silence.settings import settings
from silence.db.connector import get_conn

###############################################################################
# Reads and executes the SQL scripts needed to create the database
# and fill it with data, as defined in the settings.
###############################################################################

def create_database():
    conn = get_conn()
    cursor = conn.cursor()
    db_name = settings.DB_CONN["database"]

    cursor.execute(f"DROP DATABASE IF EXISTS `{db_name}`")
    cursor.execute(f"CREATE DATABASE `{db_name}`")
    conn.commit()
    
    for script in settings.SQL_SCRIPTS:
        cursor.execute(f"USE `{db_name}`")
        print_str = f"Executing {script}:"
        print_aux_len = len(print_str) + 2

        print("=" * print_aux_len)
        print(f" {print_str} ")
        print("=" * print_aux_len, "\n")

        with open(f"sql/{script}", "r", encoding="utf-8") as f:
            stmt = ""
            delimiter = ";"
            for line in f:
                if not line or line.strip() == "": continue  # Skip blank lines
                
                # Emulate the DELIMITER sentence
                if "DELIMITER" in line.upper():
                    delimiter = line.split(" ")[1].strip()
                else:
                    if delimiter in line.upper():
                        if delimiter != ";": stmt += line.replace(delimiter, ";")
                        else: stmt += line

                        print(stmt)
                        cursor.execute(stmt)
                        stmt = ""
                    else:
                        stmt += line

        # Commit after every SQL script
        conn.commit()
    # Clean up when we're done
    cursor.close()


import logging
import traceback
import colorama

from colorama import Fore, Style
from silence.settings import settings

if settings.COLORED_OUTPUT:
    colorama.init()

def add_color(color, format_str):
    if settings.COLORED_OUTPUT:
        return Style.BRIGHT + color + format_str + Style.RESET_ALL
    else:
        return format_str

class DefaultFormatter(logging.Formatter):

    format_nolvl = "%(message)s"
    format_lvl = "[%(levelname)s] %(message)s"

    FORMATS = {
        logging.DEBUG: add_color(Fore.BLACK, format_lvl),
        logging.INFO: format_nolvl,
        logging.WARNING: add_color(Fore.YELLOW, format_lvl),
        logging.ERROR: add_color(Fore.RED, format_lvl),
        logging.CRITICAL: add_color(Fore.RED, format_lvl)
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        res = formatter.format(record)
        
        # If we're logging an exception, print the stack trace in red
        if record.exc_info:
            res = res.replace(Style.RESET_ALL, "")
            res += Style.RESET_ALL

        return res

    



import argparse
import sys
import logging

from silence import __version__
from silence.settings import settings
from silence.logging.default_logger import logger
from silence.cli.commands import run, createdb, new, list_templates

###############################################################################
# Command line interface manager
# It provides the entry point for the 'silence' command.
###############################################################################

HANDLERS = {
    "run": run.handle,
    "createdb": createdb.handle,
    "new": new.handle,
    "list-templates": list_templates.handle
}

def run_from_command_line():
    parser = argparse.ArgumentParser(
        description="Silence: An educational framework for deploying RESTful APIs and Web applications."
    )
    subparsers = parser.add_subparsers(help="Description:", dest="command")
    
    # Force the user to select one of the available commands,
    # and allow them to provide additional options after it.
    parser.add_argument("-v", "--version", action="version", version=f"Silence v{__version__}")

    parser_list = subparsers.add_parser("list-templates", help="Lists the available project templates")
    parser_list.add_argument("--debug", action="store_true", help="Enables the debug mode")
    
    parser_new = subparsers.add_parser("new", help="Creates a new project")
    parser_new.add_argument("name", help="The new project's name")
    group = parser_new.add_mutually_exclusive_group()
    group.add_argument("--template", help="Template name to use when creating the new project")
    group.add_argument("--url", help="URL to a Git repo containing a project to clone")
    group.add_argument("--blank", action="store_true", help="Alias to --template blank")
    parser_new.add_argument("--debug", action="store_true", help="Enables the debug mode")

    parser_createdb = subparsers.add_parser("createdb", help="Runs the provided SQL scripts in the adequate order in the database")
    parser_run = subparsers.add_parser("run", help="Starts the web server")

    # Show the help dialog if the command is issued without any arguments
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    # Show silence new's help dialog if it was called without any arguments
    if sys.argv[1].lower() == "new" and len(sys.argv) == 2:
        parser_new.print_help()
        sys.exit(1)

    args = parser.parse_args()

    # If --debug is set, configure the logging level and the global settings
    # This is useful for commands that have no access to a custom settings.py
    # file, such as "new" and "list-templates"
    if "debug" in args and args.debug:
        settings.DEBUG_ENABLED = True
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    command = args.command.lower()
    HANDLERS[command](args)



from silence.settings import settings
from silence.logging.default_logger import logger

import requests
import traceback
import sys

def handle(args):
    query_url = f"https://api.github.com/orgs/{settings.GITHUB_TEMPLATES_OWNER}/repos"

    try:
        repo_data = requests.get(query_url).json()
    except Exception as exc:
        logger.debug(traceback.format_exc())
        logger.error("An error has occurred when querying GitHub's API to obtain the list of templates.")
        if not settings.DEBUG_ENABLED:
            logger.error("Add --debug to see the full stack trace.")
        sys.exit(1)

    templates = []
    for repo in repo_data:
        name = repo["name"].lower()
        if name.startswith("silence-template"):
            template_name = name.replace("silence-template-", "")
            desc = repo["description"]
            templates.append({"name": template_name, "desc": desc})

    templates.sort(key=lambda x: x["name"])

    print("Available templates:")
    for tmpl in templates:
        name = tmpl["name"]
        default = " (default)" if name == settings.DEFAULT_TEMPLATE_NAME.lower() else ""
        desc = f': {tmpl["desc"]}' if tmpl["desc"] else ""

        print(f"    · {name}{default}{desc}")



import re
import zipfile
import requests
import sys

from silence.logging.default_logger import logger

from os import listdir, remove, getcwd
from os.path import join, isfile, isdir, basename
from urllib.parse import urlparse
from shutil import move, rmtree
from secrets import token_urlsafe
from pathlib import Path

RE_REPO_URL = re.compile(r"https?:\/\/(.*?)\/(.*?)\/(.*)")

def download_from_github(project_name, repo_url):
    # Check that the current directory does not contain a folder with the same name
    if isdir(project_name):
        logger.error(f"A folder named '{project_name}' already exists in the current directory.")
        sys.exit(1)

    # Remove the trailing .git or slash if they exist
    # We could use .removesuffix, but it was added in 3.9... maybe if we bump
    # the required Python version some time in the future
    suffixes = (".git", "/")
    for suffix in suffixes:
        if repo_url.endswith(suffix):
            repo_url = repo_url[:-len(suffix)]

    # Check that the repo URL is acceptable
    m = RE_REPO_URL.match(repo_url.lower())
    if not m:
        logger.error("Invalid repo URL, please check your spelling and try again.")
        sys.exit(1)

    host, username, repo_name = m.groups()

    # Check that the host is supported
    if host not in ("github.com", "github.eii.us.es"):
        logger.error("Only repos hosted in github.com or github.eii.us.es are supported.")
        sys.exit(1)

    # Download it (this takes care of querying the relevant API to find out
    # how the default branch is called, and exiting if the repo does not exist)
    git_clone(host, username, repo_name, project_name + "/")

    # Unpack it (everything is inside the <name>-<branch> folder)
    branch_folder_name = listdir(project_name)[0]
    
    # Move everything inside that folder outside
    branch_folder = join(project_name, branch_folder_name)
    for elem in listdir(branch_folder):
        move(join(branch_folder, elem), join(project_name, elem))

    # Remove the now empty folder
    rmtree(branch_folder)

    # Look for .gitkeep files and remove them (especially useful in the case
    # of the blank template project)
    for gitkeep_path in Path(project_name).rglob(".gitkeep"):
        remove(gitkeep_path)

    # Read the settings.py file of the downloaded project, removing the existing
    # SECRET_KEY if it exists and creating a new one. Will also raise a warning
    # if the project does not contain a settings.py file.
    settings_path = join(project_name, "settings.py")

    try:
        settings_lines = open(settings_path, "r", encoding="utf-8").readlines()
        settings_lines = list(filter(lambda line: not line.startswith("SECRET_KEY"), settings_lines))

        # Generate the random string for the secret key
        # and add it to the settings.py file
        secret_key = token_urlsafe(32)

        settings_lines += [
            "\n",
            "# A random string that is used for security purposes\n",
            "# (this has been generated automatically upon project creation)\n",
            f'SECRET_KEY = "{secret_key}"\n'
        ]

        open(settings_path, "w", encoding="utf-8").writelines(settings_lines)

    except FileNotFoundError:
        logger.warning("The downloaded project does not have a settings.py file " +
         "at its root, it may not be a valid Silence project.")

def git_clone(host, username, repo_name, clone_dir):
    # Get the default branch (we've checked previously that the host is one of
    # the following)
    api_url = {
        "github.com": f"https://api.github.com/repos/{username}/{repo_name}",
        "github.eii.us.es": f"https://github.eii.us.es/api/v3/repos/{username}/{repo_name}"
    }[host]

    api_response = requests.get(api_url)
    if api_response.status_code == 404:
        logger.error("Repo not found")
        sys.exit(1)

    branch = api_response.json()["default_branch"]

    git_url = f"https://{host}/{username}/{repo_name}"
    repo_file = f"{branch}.zip"
    repo_zip = git_url + f"/archive/{repo_file}"

    if isfile(repo_file):
        remove(repo_file)

    fn = basename(urlparse(repo_zip).path)
    zip_path = save(repo_zip, join(getcwd(), fn))

    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(clone_dir)

    if isfile(zip_path):
        remove(zip_path)

# Adapted to our needs from https://github.com/x011/dload/
def save(url, path):
    r = requests.get(url)
    with open(path, 'wb') as f:
        f.write(r.content)
    return path
